#!C:\PERL\bin\perl
use warnings;
#use Benchmark; ## for benchmark purposes

#$t0 = Benchmark->new; ## Create new instace of Benchmark variable
						## begin
   

#########################################################################
## Author: Venice Margarette B. Juanillas
## Date Created: May 19, 2011
## Description: This script transforms a tabular data file into the 
## QGENE data file format
## outputs: [header]
##			Genotype Symbols:
##			A/A = 1
##			B/B = 2
##			A/B = 3						
##			C/C = 4
##			D/D = 5
##			-/- = 0
##			[locus]
##			SNP_id	cM_num	cM_pos	marker_data(123450)
##			[Trait]
##				*must still be appended; generated from another
##				*file.				
##	temporary output: out2.txt
##	final output: output.txt
#############################################################################

## declarations
my $tempf1 = "out.txt"; ## global variables that will act as temporary files for the sort
my $tempf2 = "out2.txt";
my $tempf3 = "out3.txt"; ## file to temporarily store the matrix

my $linecounter = 0; ## this variable will keep track of the lines that are already occupied


check(); 

## this subroutine will check if there are sufficient paramters issued 
sub check{  
	
	my $params = scalar(@ARGV); ##store the number of parameters
	my $string;
	my $append_how; ## prepend or append??
	
	##check for command line parameters specified
	##else prompt an error asking for files
	if($params < 3 or $params > 5){
		print "Incorrect file parameters...\n";
		print "ARGV[0] = matrix file\n"; ##user created
		print "ARGV[1] = output file\n";
		print "ARGV[2] = map file\n";	##user created
		print "ARGV[3] = header file\n";
		print "ARGV[4] = trait file\n";
	}else{
		sort_contents($ARGV[0],$tempf1); ## sort yung laman ng matrix
		sort_contents($ARGV[2],$tempf2); ## sort ang laman ng map 
		create_matrix($tempf1,$tempf2,$tempf3); ## creates the matrix in another file
		
		## check if header and trait files exist
		if(!$ARGV[3] or !$ARGV[4]){
			if(!$ARGV[3]){
				print "No header section\n";
				$string = "[Header]"."\n"."[Locus]";
				$append_how = "+>";	## prepend symbol
				append_string($append_how.$ARGV[1],$string);
			}
			
			append_matrix($ARGV[1],$tempf3); ## append the matrix
			
			if(!$ARGV[4]){
				print "No trait section\n";
				$string = "[Trait]";
				$append_how = "+>>"; ## append symbol
				append_string($append_how.$ARGV[1],$string);
			}
		}else{
			append_header($ARGV[3],$ARGV[1]); ##subroutine calls
			append_matrix($ARGV[1],$tempf3);
			append_trait($ARGV[1],$ARGV[4]);
		}
	}
}

## this subroutine will read the matrix from the temporary file
## and append it to the output file
sub append_matrix{
	my ($output,$matrix) = @_; 
	my $line;
	my $linecount = 1;
	my $to_string;
	my @row;
	my @lines;
	
	## open files
	open(IN,"$matrix")|| die "Cannot open $matrix";
	open(OUT,"+>>$output") || die "Cannot open $output";
	
	while($line = <IN>){
		next if $line=~ /^\s*$/; ## skip if blank line
		if($linecount == 1){
			$linecount++;
			next;
		}else{
			@row = split(/\t/,$line); ## split row 
			($row[0],$row[1])= ($row[1],$row[0]); ## swap cM num and SNP
			$to_string = join("\t",@row);
			print OUT $to_string; ## append row by row
		}
	}
	
	##close files
	close(IN);
	close(OUT);
}

## this subroutine will create the matrix of the locus section
## however, this will create the matrix in a temporary file
sub create_matrix{
	my ($input1,$input2,$output) = @_; ## input1: matrix input2: map input3: temp_out
	my $line;
	my $linecount = 1; ## flag if header
	my $count = 0;
	my $string;
	my $hit = 0; ## flag if SNP id found
	my @row;
	my @markers;
	my @SNP_rows;
	my $alleles;
	
	open(IN1,"$input1") || die "Cannot open $input1";
	open(OUT,">$output") || die "Cannot open $output";
	
	while($line = <IN1>){
		next if $line =~ /^\s*$/; ## we skip the blank lines
		
		## skip the headers
		if($linecount == 1){
			$linecount++;
			next;
		}elsif($linecount ge 1){
			@row = split(/\t/,$line);
			$hit = search_SNPid($row[0],$input2); ## search if SNP in matrix has mappings
			$alleles = get_alleles($row[0],$input1); 
		
		## if SNP is has a cM mapping
			if($hit == 1){				
				$string = get_mappings($row[0],$input2); ## get the chromosome mappings
			}else{
				
				$string = "-"."\t".$row[0]."\t"."-"."\t"; ## if there exist no mappings,cM & cM_pos= "-"
			}
			$SNP_rows[$count] = $string.$alleles."\n";	
			$linecount++;
			$count++;
		}
	}
	## sort the SNPs by their chromosomes
	@SNP_rows = sort(@SNP_rows);
	
	## print SNPs in temporary file
	for($i = 0; $i < $#SNP_rows; $i++){
		print OUT $SNP_rows[$i];
	}
	
	close(IN1);
	close(OUT);
}

## this will form a string that comprise only of the alleles of a certain SNP id
sub get_alleles{
	my ($SNP, $file) = @_;
	my $line;
	my $linecount = 1; ## flag if we are in the header row
	my @row;
	my @markers; ## store here all marker data
	my $str;
	
	open(IN,"$file") || die "Cannot open $file.";
	
	while($line = <IN>){
		next if $line =~ /^\s*$/;
		if($linecount == 1){ ## skip the header
			$linecount++;
			next;
		}
		if($linecount != 1){
			@row= split(/\t/,$line); ## split whole row
			if($row[0] eq $SNP){	## we change the genotype symbols from ABHCDx : 1234560
				for($i = 1; $i < $#row; $i++){
					if ($row[$i] eq "A/A"){ 
						$row[$i] = "1"
					}
					if ($row[$i] eq "B/B"){
						$row[$i] = "2"
					}
					if ($row[$i]eq "A/B"){
						$row[$i] = "3"
					}
					if ($row[$i] eq "C/C"){
						$row[$i] = "4"
					}
					if ($row[$i] eq "D/D"){
						$row[$i] = "5"
					}
					if ($row[$i] eq "-/-"){
						$row[$i] = "6"
					}
					push(@markers,$row[$i]); ## push all into an array
				}
			}
		}
	}
	$str = join("\t",@markers); ## delimit the markers by tabs
	
	close(IN);
	return $str; ## return the string
}

## this subroutine will get the cM, cM position and the SNP ids
sub get_mappings{
	my ($SNP,$file) = @_;
	my $line;
	my $mappings;
	my $linecount = 1; ## flag if header
	my @row;
	
	open(IN,$file) || die $!;
	
	while($line = <IN>){
		next if $line =~ /^\s*$/; ## we skip the blank lines
		if($linecount == 1){
			$linecount++;
			next;
		}else{
			@row = split(/\t/,$line);
			if($row[0] eq $SNP){
				$mappings = $row[1]."\t".$row[0]."\t".$row[2]."\t";
				return $mappings;
			}
			$linecount++;
		}
	}
	
	close(IN);
}

## this subroutine will append the marker data 
## in the matrix
sub change_marker{
	my ($marker) = @_;
	
	if ($marker eq "A/A"){ 
		$marker = "1"
	}
	if ($marker eq "B/B"){
		$marker = "2"
	}
	if ($marker eq "A/B"){
		$marker = "3"
	}
	if ($marker eq "C/C"){
		$marker = "4"
	}
	if ($marker eq "D/D"){
		$marker = "5"
	}
	if ($marker eq "-/-"){
		$marker = "6"
	}
					
	return $marker; ## return the string
}

## this subroutine checks if the SNP is in the cM mapping
## drawback: this will be slow...
## what if millions of SNPs? very, very slow
## Solution: Search for the fastest search algo....
sub search_SNPid{
	my ($to_search,$file_to_search )= @_;
	my $line;
	my $flag = 0; ## mark if the SNP id exists in the map
	my $linecount = 1;
	
	open(FILE,"$file_to_search") || die "Cannot open $file_to_search.";
	
	while($line = <FILE>){
		next if $line =~ /^\s*$/;
		if($linecount == 1){ ## we skip the headers
			$linecount++;
		}
		elsif($linecount != 1){
			@row= split(/\t|\n/,$line);
			if($row[0] eq $to_search){ ## compare with the id in the map
				$flag = 1;
			}
		}
	}
	close(FILE);
	
	return $flag;

}

## this subroutine will sort its contents based on SNPs
## for faster searching 
sub sort_contents{
	my ($infile,$outfile) = @_;
	my @array;
	
	## open the files
	open(IN,"$infile") || die "Cannot open $infile";
	open(OUT,"+>$outfile") || die "Cannot open $outfile"; ## we keep the sorted contents in the temporary files
	
	@array = <IN>;
	@array = sort(@array); ## sort
	print OUT @array; ## then output all 

	## close to save
	close(IN);
	close(OUT);
}

## this will append any given string to a given file 
sub append_string{
	my ($file, $string) = @_;
	my $line;
	
	open(IN,"$file") || die "Cannot open $file"; ## open file
	print IN $string."\n"; ##write/append to the file
	close(IN); ## close
}


## this subroutine will get the header from header.txt 
## and write it to the output file
sub append_header{
	my ($infile,$outfile) = @_;
	my $line;
	
	## open files
	open(IN,"$infile") || die "Cannot open file.";
	open(OUT,">$outfile") || die "Cannot open file.";
	
	
	##traverse through the input file
	while($line = <IN>){
		next if $line =~ /^\s*$/;
		print OUT $line;
		$linecounter++;
	}	
	
	print OUT "\n";
	## close all files to save modifications
	close(IN);
	close(OUT);

}

## this subroutine will append the Trait Section
sub append_trait{
	my ($file,$trait_file) = @_;
	my $line;
	
	##open file for manipulation
	open(FILE,"+>>$file") || die "Cannot open $file";
	open(F,"+<$trait_file") || die "Cannot open $file";
	
	## read contents of trait file and copy to the output file
	while($line = <F>){
		print FILE $line;
	}
	
	##close files to save all modifications
	close(FILE);
	close(F);
}

#$t1 = Benchmark->new;
#$td = timediff($t1, $t0);
#print "the code took:",timestr($td),"\n";



## end of the script
## Date Completed: May 23,2011
## Revised: May 28,2011
## Remarks: For further testing

